@@ -21,7 +21,8 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
-#include <fstream.h>
+
+#include <fstream>
 
 extern "C"
 {
@@ -41,10 +42,23 @@ extern "C"
 #include "flag_opt_arg.h"
 #include "required_option.h"
 #include "generic_option.h"
+#include "group_option.h"
+#include "group_counter.h"
+#include "generic_option_group.h"
 #include "handle_help.h"
 #include "handle_version.h"
 #include "no_short_option.h"
 #include "handle_unamed.h"
+#include "print_group.h"
+#include "config_parser_header.h"
+#include "multiple_opt_list.h"
+#include "multiple_option.h"
+#include "multiple_fill_array.h"
+
+#include "conf_parser_gen.h"
+
+using std::ofstream;
+using std::endl;
 
 extern char * gengetopt_package;
 extern char * gengetopt_version;
@@ -52,13 +66,12 @@ extern char * gengetopt_purpose;
 extern int gengetopt_strdup_text_length;
 extern char *gengetopt_strdup_text[];
 
+extern groups_collection_t gengetopt_groups;
+
 static char *create_filename (char *name, char *ext);
 static ofstream *open_fstream (char *filename);
 static char *canonize_names(char * name);
 
-#define ARGS_STRUCT "args_info"
-#define ARGS_STRUCT_NAME PACKAGE "_" ARGS_STRUCT
-
 char *
 create_filename (char *name, char *ext)
 {
@@ -99,12 +112,14 @@ CmdlineParserCreator::CmdlineParserCreator (char *function_name,
                                             bool long_help_, 
                                             bool no_handle_help_,
                                             bool no_handle_version_,
-                                            bool no_handle_error_, 
+                                            bool no_handle_error_,
+                                            bool conf_parser_,
                                             const string &comment_) :
   filename (filename_), comment (comment_),
   unamed_options (unamed_options_), 
   long_help (long_help_), no_handle_help (no_handle_help_),
   no_handle_version (no_handle_version_), no_handle_error (no_handle_error_),
+  conf_parser (conf_parser_),
   tab_indentation (0)
 {
   parser_function_name = canonize_names (function_name);
@@ -121,6 +136,18 @@ CmdlineParserCreator::CmdlineParserCreator (char *function_name,
   if (gengetopt_version)
     set_if_version_notdefined (gengetopt_version);
 
+  if (conf_parser)
+    {
+      // the config file parser prototype
+      ostringstream conf_parser_proto;
+      config_parser_header_gen_class cf_gen;
+      cf_gen.set_parser_name (parser_function_name);
+      conf_parser_proto << endl;
+      cf_gen.generate_config_parser_header (conf_parser_proto);
+      conf_parser_proto << endl;
+      set_config_parser_proto (conf_parser_proto.str ());
+    }
+
   // c_source_gen_class
   set_command_line (comment);
   c_source_gen_class::set_generator_version (VERSION);
@@ -145,12 +172,32 @@ CmdlineParserCreator::do_update_arg (struct gengetopt_option *opt,
                                      ostream &stream,
                                      unsigned int indent) 
 {
+  const string argstr ("optarg");
   string_opt_arg_gen_class str_gen;
+  str_gen.set_optarg (argstr);
   int_opt_arg_gen_class int_gen;
+  int_gen.set_optarg (argstr);
   long_opt_arg_gen_class long_gen;
+  long_gen.set_optarg (argstr);
   float_opt_arg_gen_class float_gen;
+  float_gen.set_optarg (argstr);
   flag_opt_arg_gen_class flag_gen;
 
+  if (opt->multiple)
+    {
+      int_gen.set_structure (string (opt->var_arg) + "_new");
+      str_gen.set_structure (string (opt->var_arg) + "_new");
+      long_gen.set_structure (string (opt->var_arg) + "_new");
+      float_gen.set_structure (string (opt->var_arg) + "_new");
+    }
+  else
+    {
+      int_gen.set_structure (ARGS_STRUCT);
+      str_gen.set_structure (ARGS_STRUCT);
+      long_gen.set_structure (ARGS_STRUCT);
+      float_gen.set_structure (ARGS_STRUCT);
+    }
+
   switch (opt->type) 
     {
     case ARG_NO:
@@ -273,7 +320,11 @@ CmdlineParserCreator::generate_option_arg(ostream &stream,
         else
           first = false;
 
-        stream << arg_types[opt->type] << " ";
+        stream << arg_types[opt->type];
+	if (opt->multiple)
+	  stream << "* ";
+	else
+	  stream << " ";
         break;
       default: fprintf (stderr, "gengetopt: bug found in %s:%d!!\n",
                         __FILE__, __LINE__);
@@ -289,10 +340,7 @@ CmdlineParserCreator::generate_option_arg(ostream &stream,
       
       if (opt->default_given)
         {
-          if (opt->type == ARG_STRING)
-            stream << " (default='" << opt->default_string << "')";
-          else
-            stream << " (default=" << opt->default_num << ")";
+          stream << " (default='" << opt->default_string << "')";
         }
       
       if (opt->type == ARG_FLAG)
@@ -492,8 +540,22 @@ CmdlineParserCreator::generate_help_option_print(ostream &stream,
       }
   }
   /* print justified options */
+  char *prev_group = 0;
+  print_group_gen_class print_group;
+
   foropt
     {
+      if (opt->group_value && 
+          (! prev_group || strcmp (opt->group_value, prev_group) != 0))
+        {
+          stream << indent_str;
+          print_group.set_group_name (opt->group_value);
+          print_group.generate_print_group (stream, indent);
+          stream << endl;
+
+          prev_group = opt->group_value;
+        }
+
       stream << indent_str;
       stream << "printf(\"";
       if (opt->type == ARG_FLAG || opt->type == ARG_NO)
@@ -538,10 +600,7 @@ CmdlineParserCreator::generate_help_option_print(ostream &stream,
           stream << "  " << opt->desc;
           if (opt->default_given)
             {
-              if (opt->type == ARG_STRING)
-                stream << " (default='" << opt->default_string << "')";
-              else
-                stream << " (default=" << opt->default_num << ")";
+              stream << " (default='" << opt->default_string << "')";
             }
           stream << "\\n\");\n";
         }
@@ -589,6 +648,55 @@ CmdlineParserCreator::generate_given_init(ostream &stream,
     }
 }
 
+void
+CmdlineParserCreator::generate_struct_def(ostream &stream, unsigned int indent)
+{
+  struct gengetopt_option * opt;
+  string indent_str (indent, ' ');
+  multiple_opt_list_gen_class multiple_opt;
+  bool generated_counter = false;
+
+  /* define linked-list structs for multiple options */
+  foropt
+    {
+      if (opt->multiple)
+	{
+          if (! generated_counter)
+            {
+              stream << indent_str;
+              stream << "int i;        /* Counter */" << endl;
+              generated_counter = true;
+            }
+          stream << indent_str;
+	  multiple_opt.set_type (arg_types[opt->type]);
+	  multiple_opt.set_arg_name (opt->var_arg);
+	  multiple_opt.generate_multiple_opt_list (stream, indent);
+          stream << endl;
+	}
+    }
+}
+
+void
+CmdlineParserCreator::generate_multiple_fill_array(ostream &stream, unsigned int indent)
+{
+  struct gengetopt_option * opt;
+  string indent_str (indent, ' ');
+  multiple_fill_array_gen_class filler;
+
+  /* copy linked list into the array */
+  foropt
+    {
+      if (opt->multiple)
+	{
+	  stream << indent_str;
+	  filler.set_type (arg_types[opt->type]);
+	  filler.set_option_var_name (opt->var_arg);
+	  filler.generate_multiple_fill_array (stream, indent);
+          stream << endl;
+	}
+    }
+}
+
 void
 CmdlineParserCreator::generate_clear_arg(ostream &stream, unsigned int indent)
 {
@@ -626,7 +734,7 @@ CmdlineParserCreator::generate_clear_arg(ostream &stream, unsigned int indent)
           stream << indent_str;
 
           stream << ARGS_STRUCT << "->" << opt->var_arg << "_arg";
-          stream << " = " << opt->default_num << " ;\\";
+          stream << " = " << opt->default_string << " ;\\";
 
           stream << endl;
         }
@@ -785,6 +893,7 @@ CmdlineParserCreator::generate_handle_option(ostream &stream,
 {
   struct gengetopt_option * opt;
   generic_option_gen_class option_gen;
+  multiple_option_gen_class multi_gen;
   string indent_str (indent, ' ');
 
   if (no_handle_error)
@@ -802,13 +911,29 @@ CmdlineParserCreator::generate_handle_option(ostream &stream,
           stream << indent_str;
 
           string short_opt (1, opt->short_opt);
+	  if (opt->multiple)
+	    {
+	      multi_gen.set_short_option (short_opt);
+	      multi_gen.set_option_comment (opt->desc);
+	      multi_gen.set_option_var_name (opt->var_arg);
 
-          option_gen.set_short_option (short_opt);
-          option_gen.set_option_comment (opt->desc);
-          option_gen.set_long_option (opt->long_opt);
-          option_gen.set_option_var_name (opt->var_arg);
+	      multi_gen.generate_multiple_option (stream, indent);
+	    }
+	  else
+	    {
+	      option_gen.set_short_option (short_opt);
+	      option_gen.set_option_comment (opt->desc);
+	      option_gen.set_long_option (opt->long_opt);
+	      option_gen.set_option_var_name (opt->var_arg);
 
-          option_gen.generate_generic_option (stream, indent);
+	      option_gen.generate_generic_option (stream, indent);
+	    }
+          generic_option_group_gen_class group_gen;
+          if (opt->group_value)
+            {
+              group_gen.set_group_var_name (opt->group_value);
+              group_gen.generate_generic_option_group (stream, indent);
+            }
 
           do_update_arg (opt, stream, indent + 2);
 
@@ -818,6 +943,40 @@ CmdlineParserCreator::generate_handle_option(ostream &stream,
     }
 }
 
+#define GROUP_REQUIRED_COMPARISON "!="
+#define GROUP_NOT_REQUIRED_COMPARISON ">"
+#define GROUP_REQUIRED_MESSAGE "One"
+#define GROUP_NOT_REQUIRED_MESSAGE "At most one"
+
+void
+CmdlineParserCreator::generate_handle_group(ostream &stream,
+                                            unsigned int indent)
+{
+  group_option_gen_class opt_gen;
+  string indent_str (indent, ' ');
+
+  groups_collection_t::const_iterator end = gengetopt_groups.end();
+  for ( groups_collection_t::const_iterator idx = gengetopt_groups.begin(); 
+        idx != end; ++idx)
+    {
+      stream << indent_str;
+      opt_gen.set_group_var_name (idx->first);
+      if (idx->second)
+        {
+          opt_gen.set_Comparison_rule(GROUP_REQUIRED_COMPARISON);
+          opt_gen.set_number_required(GROUP_REQUIRED_MESSAGE);
+        }
+      else
+        {
+          opt_gen.set_Comparison_rule(GROUP_NOT_REQUIRED_COMPARISON);
+          opt_gen.set_number_required(GROUP_NOT_REQUIRED_MESSAGE);
+        }
+		
+      opt_gen.generate_group_option (stream, indent);
+      stream << endl;
+    }
+}
+
 void
 CmdlineParserCreator::generate_handle_required(ostream &stream, 
                                                unsigned int indent)
@@ -846,6 +1005,24 @@ CmdlineParserCreator::generate_handle_required(ostream &stream,
       }
 }
 
+void
+CmdlineParserCreator::generate_group_counters(ostream &stream, 
+                                              unsigned int indent)
+{
+  group_counter_gen_class counter_gen;
+  string indent_str (indent, ' ');
+
+  groups_collection_t::const_iterator end = gengetopt_groups.end();
+  for ( groups_collection_t::const_iterator idx = gengetopt_groups.begin();
+        idx != end; ++idx)
+    {
+      stream << indent_str;
+      counter_gen.set_group_name (idx->first);
+      counter_gen.generate_group_counter (stream, indent);
+      stream << endl;
+    }     
+}
+
 int
 CmdlineParserCreator::generate_source ()
 {
@@ -877,6 +1054,20 @@ CmdlineParserCreator::generate_handle_unamed(ostream &stream,
   stream << endl;
 }
 
+void
+CmdlineParserCreator::generate_conf_parser(ostream &stream, 
+                                           unsigned int indent)
+{
+  if (! conf_parser)
+    return;
+
+  stream << endl;
+  ConfigParserGenerator conf_parser_gen;
+  conf_parser_gen.set_parser_name (c_source_gen_class::parser_name);
+  conf_parser_gen.set_gen_exit (gen_exit);
+  conf_parser_gen.generate_config_parser_source (stream);
+}
+
 /*
   return a copy of the string passed after canonizing it (i.e. '-' and
   '.' are transformed in '_'.
