@@ -21,11 +21,16 @@
 #ifndef _GM_H
 #define _GM_H
 
+#include "my_string.h"
+#include "my_map.h"
+
 #include "header.h"
 #include "c_source.h"
 
 #define TAB_LEN 2
 
+typedef map<string,bool> groups_collection_t;
+
 struct gengetopt_option;
 
 class CmdlineParserCreator : public header_gen_class, public c_source_gen_class
@@ -42,6 +47,7 @@ class CmdlineParserCreator : public header_gen_class, public c_source_gen_class
   bool no_handle_help;
   bool no_handle_version;
   bool no_handle_error;
+  bool conf_parser;
 
   unsigned int tab_indentation ; /* tab indentation level */
 
@@ -80,12 +86,16 @@ class CmdlineParserCreator : public header_gen_class, public c_source_gen_class
   virtual void generate_print_purpose(ostream &stream, unsigned int indent);
   virtual void generate_clear_arg(ostream &stream, unsigned int indent);
   virtual void generate_given_init(ostream &stream, unsigned int indent);
+  virtual void generate_struct_def(ostream &stream, unsigned int indent);
+  virtual void generate_multiple_fill_array(ostream &stream, unsigned int indent);
   virtual void generate_init_unamed(ostream &stream, unsigned int indent);
   virtual void generate_handle_help(ostream &stream, unsigned int indent);
   virtual void generate_handle_no_short_option(ostream &stream, 
                                                unsigned int indent);
   virtual void generate_handle_option(ostream &stream, unsigned int indent);
   virtual void generate_handle_required(ostream &stream, unsigned int indent);
+  virtual void generate_handle_group(ostream &stream, unsigned int indent);
+  virtual void generate_group_counters(ostream &stream, unsigned int indent);
   virtual void generate_handle_version(ostream &stream, unsigned int indent);
   virtual void generate_help_option_print(ostream &stream, 
                                           unsigned int indent);
@@ -93,6 +103,7 @@ class CmdlineParserCreator : public header_gen_class, public c_source_gen_class
                                            unsigned int indent);
   virtual void generate_strdup(ostream &stream, unsigned int indent);
   virtual void generate_handle_unamed(ostream &stream, unsigned int indent);
+  virtual void generate_conf_parser(ostream &stream, unsigned int indent);
 
   void generateBreak(ostream &stream, unsigned int indent);
 
@@ -102,7 +113,8 @@ class CmdlineParserCreator : public header_gen_class, public c_source_gen_class
                         char *filename, char *header_ext, char *c_ext,
                         bool long_help, bool no_handle_help,
                         bool no_handle_version,
-                        bool no_handle_error, const string &comment);
+                        bool no_handle_error, bool conf_parser,
+                        const string &comment);
 
   int generate();
 };
